# -*- coding: utf-8 -*-
from ..logger import logger
logger.debug("Started 'reading app/plugin_manager.py'")

from glob import glob
import os, sys, difflib, zipfile, time, shutil, traceback
from os.path import expanduser
from qtpy import QtGui, QtWidgets, QtCore
from urllib.request import urlopen
from urllib.error import HTTPError
from urllib.parse import urljoin
import threading
import tempfile
from xml.etree import ElementTree
import platform
import pkg_resources

from .. import global_vars as g
from ..utils.misc import load_ui
from ..images import image_path

plugin_list = {
    'Global Analysis':  'https://raw.githubusercontent.com/BrettJSettle/GlobalAnalysisPlugin/master/',
    'Beam Splitter':    'https://raw.githubusercontent.com/BrettJSettle/BeamSplitter/master/',
    'Rodent Tracker':   'https://raw.githubusercontent.com/kyleellefsen/rodentTracker/master/',
    'Detect Puffs':     'https://raw.githubusercontent.com/kyleellefsen/detect_puffs/master/',
    'Pynsight':         'http://raw.githubusercontent.com/kyleellefsen/pynsight/master/'
}

helpHTML = '''
<h1 style="width:100%; text-align:center">Welcome to the flika Plugin Manager</h1>
<p>Use the search bar to the left to find a specific plugin, or browse the list below it.</p>

<div>
    <h3>Develop a new plugin</h3>
    <p> If you would like to develop your own plugin for flika, follow these simple steps:</p>
    <ul list-style=none>
        <li>1. Download <a href="https://github.com/flika-org/flika_plugin_template">flika Plugin Template</a> and place it in your .FLIKA/plugins directory</li>
        <li>2. Update the info.xml file for your plugin</li>
        <li>3. Refer to the <a href="http://flika-org.github.io/writing_plugins.html">flika Documentation</a> for assistance developing your plugin.</li>
        <li>4. Update the description.html file for your plugin</li>
        <li>5. Send your plugin repo to us and we'll add it to the Plugin Manager!</li>
    </ul>
<div>
'''


def get_plugin_directory():
    logger.debug('Calling app.plugin_manager.get_plugin_directory')
    local_flika_directory = os.path.join(expanduser("~"), '.FLIKA')
    plugin_directory = os.path.join(expanduser("~"), '.FLIKA', 'plugins' )
    if not os.path.exists(plugin_directory):
        os.makedirs(plugin_directory)
    if not os.path.isfile(os.path.join(plugin_directory, '__init__.py')):
        open(os.path.join(plugin_directory, '__init__.py'), 'a').close()  # Create empty __init__.py file
    if plugin_directory not in sys.path:
        sys.path.append(plugin_directory)
    if local_flika_directory not in sys.path:
        sys.path.append(local_flika_directory)
    return plugin_directory

plugin_dir = get_plugin_directory()


def parse(x):
    #logger.debug('Calling app.plugin_manager.parse')
    tree = ElementTree.fromstring(x)
    def step(item):
        d = {}
        if item.text and item.text.strip():
            d['#text'] = item.text.strip()
        for k, v in item.items():
            d['@%s' % k] = v
        for k in item.getchildren():
            if k.tag not in d:
                d[k.tag] = step(k)
            elif type(d[k.tag]) == list:
                d[k.tag].append(step(k))
            else:
                d[k.tag] = [d[k.tag], step(k)]
        if len(d) == 1 and '#text' in d:
            return d['#text']
        return d
    return step(tree)


def str2func(plugin_name, file_location, function):
    '''
    takes plugin_name, path to object, function as arguments
    imports plugin_name.path and gets the function from that imported object
    to be run when an action is clicked
    '''
    #logger.debug("Started 'app.plugin_manager.str2func({}, {}, {})'".format(plugin_name, file_location, function))
    __import__(plugin_name)
    plugin_dir = "plugins.{}.{}".format(plugin_name, file_location)
    levels = function.split('.')
    module = __import__(plugin_dir, fromlist=[levels[0]]).__dict__[levels[0]]
    for i in range(1, len(levels)):
        module = getattr(module, levels[i])
    #logger.debug("Completed 'app.plugin_manager.str2func({}, {}, {})'".format(plugin_name, file_location, function))
    return module


def fake_str2func(plugin_name, file_location, function):
    def fake_fun():
        print(str(function))
        print('yay')
    return fake_fun

def build_submenu(module_name, parent_menu, layout_dict):
    #logger.debug('Calling app.plugin_manager.build_submenu')
    if len(layout_dict) == 0:
        g.alert("Error building submenu for the plugin '{}'. No items found in 'menu_layout' in the info.xml file.".format(module_name))
    for key, value in layout_dict.items():
        if type(value) != list:
            value = [value]
        if key == 'menu':
            for v in value:
                menu = parent_menu.addMenu(v["@name"])
                build_submenu(module_name, menu, v)
        elif key == 'action':
            for od in value:
                method = str2func(module_name, od['@location'], od['@function'])
                if method is not None:
                    action = QtWidgets.QAction(od['#text'], parent_menu, triggered = method)
                    parent_menu.addAction(action)


class Plugin():
    def __init__(self, name=None, info_url=None):
        self.name = name
        self.directory = None
        self.url = None
        self.author = None
        self.documentation = None
        self.version = ''
        self.latest_version = ''
        self.menu = None
        self.listWidget = QtWidgets.QListWidgetItem(self.name)
        self.installed = False
        self.description = ''
        self.dependencies = []
        self.loaded = False
        self.info_url = info_url
        if info_url:
            self.update_info()

    def lastModified(self):
        return os.path.getmtime(os.path.join(plugin_dir, self.directory))

    def fromLocal(self, path):
        #logger.debug('Calling app.plugin_manager.Plugin.fromLocal')
        text = open(os.path.join(path, 'info.xml'), 'r').read()
        info = parse(text)
        self.name = info['@name']
        self.directory = info['directory']
        self.version = info['version']
        self.latest_version = self.version
        self.author = info['author']
        try:
            self.description = str(open(os.path.join(path, 'about.html'), 'r').read())
        except FileNotFoundError:
            self.description = "No local description file found"
        self.url = info['url'] if 'url' in info else None
        self.documentation = info['documentation'] if 'documentation' in info else None
        if 'dependencies' in info and 'dependency' in info['dependencies']:
            deps = info['dependencies']['dependency']
            self.dependencies = [d['@name'] for d in deps] if isinstance(deps, list) else [deps['@name']]
        self.menu_layout = info.pop('menu_layout')
        self.listWidget = QtWidgets.QListWidgetItem(self.name)
        self.listWidget.setIcon(QtGui.QIcon(image_path('check.png')))
        self.loaded = True

    def bind_menu_and_methods(self):
        if len(self.menu_layout) > 0:
            self.menu = QtWidgets.QMenu(self.name)
            build_submenu(self.directory, self.menu, self.menu_layout)
        else:
            self.menu = None


    def update_info(self):
        logger.debug('Calling app.plugin_manager.update_info')
        if self.info_url is None:
            return False
        info_url = urljoin(self.info_url, 'info.xml')
        try:
            txt = urlopen(info_url).read()
        except HTTPError as e:
            g.alert("Failed to update information for {}.\n\t{}".format(self.name, e))
            return

        new_info = parse(txt)
        description_url = urljoin(self.info_url, 'about.html')
        try:
            new_info['description'] = urlopen(description_url).read().decode('utf-8')
        except HTTPError:
            new_info['description'] = "Unable to get description for {0} from <a href={1}>{1}</a>".format(self.name, description_url)
        self.menu_layout = new_info.pop('menu_layout')
        if 'date' in new_info:
            new_info['version'] = '.'.join(new_info['date'].split('/')[2:] + new_info['date'].split('/')[:2])
            new_info.pop('date')
        new_info['latest_version'] = new_info.pop('version')
        if 'dependencies' in new_info and 'dependency' in new_info['dependencies']:
            deps = new_info.pop('dependencies')['dependency']
            self.dependencies = [d['@name'] for d in deps] if isinstance(deps, list) else [deps['@name']]
        self.__dict__.update(new_info)
        self.loaded = True
        

class PluginManager(QtWidgets.QMainWindow):
    plugins = {}
    loadThread = None
    sigPluginLoaded = QtCore.Signal(str)
    '''
    PluginManager handles installed plugins and the online plugin database
    | show() : initializes a gui as a static variable of the class, if necessary, and displays it. Call in place of constructor
    | close() : closes the gui if it exists
    '''
    @staticmethod
    def show():
        logger.debug('Calling app.plugin_manager.PluginManager.show')
        if not hasattr(PluginManager, 'gui'):
            PluginManager.gui = PluginManager()
        PluginManager.gui.showPlugins()
        #PluginManager.load_online_plugins()
        QtWidgets.QMainWindow.show(PluginManager.gui)
        if not os.access(plugin_dir, os.W_OK):
            g.alert("Plugin folder write permission denied. Restart flika as administrator to enable plugin installation.")

        PluginManager.gui.showHelpScreen()


    @staticmethod
    def refresh_online_plugins():
        logger.debug('Calling app.plugin_manager.PluginManager.refresh_online_plugins()')
        for p in plugin_list.keys():
            PluginManager.load_online_plugin(p)

    @staticmethod
    def load_online_plugin(p):
        logger.debug('Calling app.plugin_manager.PluginManager.load_online_plugin()')
        if p not in plugin_list or PluginManager.loadThread is not None and PluginManager.loadThread.is_alive():
            return
        def loadThread():
            plug = PluginManager.plugins[p]
            plug.info_url = plugin_list[p]
            plug.update_info()
            PluginManager.gui.sigPluginLoaded.emit(p)
            #PluginManager.gui.statusBar.showMessage('Plugin information loaded successfully')

        PluginManager.loadThread = threading.Thread(None, loadThread)
        PluginManager.gui.statusBar.showMessage('Loading plugin information for {}...'.format(p))
        PluginManager.loadThread.start()

    def closeEvent(self, ev):
        if self.loadThread is not None and self.loadThread.is_alive():
            self.loadThread.join(0)

    @staticmethod
    def close():
        if hasattr(PluginManager, 'gui'):
            QtWidgets.QMainWindow.close(PluginManager.gui)

    def __init__(self):
        logger.debug('Calling app.plugin_manager.PluginManager.load_online_plugin()')

        super(PluginManager,self).__init__()
        load_ui("plugin_manager.ui", self, directory=os.path.dirname(__file__))
        try:
            self.scrollAreaWidgetContents.setContentsMargins(10, 10, 10, 10)
        except:
            pass
        #self.pluginList.itemClicked.connect(self.pluginSelected)
        self.tutorialButton.clicked.connect(lambda : QtGui.QDesktopServices.openUrl(QtCore.QUrl("https://github.com/flika-org/flika_plugin_template")))
        self.open_plugins_directory_button.clicked.connect(lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl("file:///" + os.path.expanduser('~/.FLIKA/plugins/'))))

        self.downloadButton.clicked.connect(self.downloadClicked)
        self.pluginList.currentItemChanged.connect(lambda new, old: self.pluginSelected(new))
        self.documentationButton.clicked.connect(self.documentationClicked)
        self.updateButton.clicked.connect(self.updateClicked)
        
        self.searchBox.textChanged.connect(self.showPlugins)
        self.searchButton.clicked.connect(lambda f: self.showPlugins(search_str=str(self.searchBox.text())))
        self.descriptionLabel.setOpenExternalLinks(True)
        
        self.refreshButton.pressed.connect(self.refresh_online_plugins)
        def updatePlugin(a):
            self.statusBar.showMessage("Finished loading {}".format(a))
            if PluginManager.plugins[a].listWidget.isSelected():
                PluginManager.gui.pluginSelected(a)
            #else:
                #self.showPlugins()
        self.sigPluginLoaded.connect(updatePlugin)

        self.setWindowTitle('Plugin Manager')
        self.showPlugins()

    def showHelpScreen(self):
        self.pluginLabel.setText("")
        self.descriptionLabel.setHtml(helpHTML)
        self.downloadButton.setVisible(False)
        self.documentationButton.setVisible(False)
        self.updateButton.setVisible(False)
        self.infoLabel.setText("")

    def downloadClicked(self):
        p = str(self.pluginList.currentItem().text())
        plugin = self.plugins[p]
        if self.downloadButton.text() == 'Install':
            PluginManager.downloadPlugin(plugin)
        else:
            PluginManager.removePlugin(plugin)

    def documentationClicked(self):
        p = str(self.pluginList.currentItem().text())
        plugin = self.plugins[p]
        if hasattr(plugin, "documentation"):
            QtGui.QDesktopServices.openUrl(QtCore.QUrl(plugin.documentation))

    def updateClicked(self):
        p = str(self.pluginList.currentItem().text())
        plugin = self.plugins[p]
        PluginManager.removePlugin(plugin)
        PluginManager.downloadPlugin(plugin)

    def pluginSelected(self, item):
        from pkg_resources import parse_version
        logger.debug('Calling app.plugin_manager.PluginManager.pluginSelected()')
        if item is None:
            if self.pluginLabel.text():
                self.pluginSelected(PluginManager.plugins[self.pluginLabel.text()].listWidget)
            return
        if isinstance(item, str):
            s = item
        else:
            s = str(item.text())
        plugin = self.plugins[s]
        self.pluginLabel.setText(s)
        if not plugin.loaded:
            info = "Loading information"
        else:
            info = 'By {}, Latest: {}'.format(plugin.author, plugin.latest_version)
            self.downloadButton.setVisible(True)
        version = parse_version(plugin.version)
        latest_version = parse_version(plugin.latest_version)
        if plugin.version and version < latest_version:
            info += "; <b>Update Available!</b>"

        self.updateButton.setVisible(plugin.version != '' and version < latest_version)
        self.downloadButton.setText("Install" if plugin.version == '' else 'Uninstall')
        self.documentationButton.setVisible(plugin.documentation != None)
        if plugin.version == '':
            plugin.listWidget.setIcon(QtGui.QIcon())
        elif parse_version(plugin.version) < parse_version(plugin.latest_version):
            plugin.listWidget.setIcon(QtGui.QIcon(image_path('exclamation.png')))
        else:
            plugin.listWidget.setIcon(QtGui.QIcon(image_path('check.png')))

        self.infoLabel.setText(info)
        self.descriptionLabel.setHtml(plugin.description)
        if plugin.info_url == None:
            self.load_online_plugin(plugin.name)

    @staticmethod
    def local_plugin_paths():
        paths = []
        for path in glob(os.path.join(plugin_dir, "*")):
            if os.path.isdir(path) and os.path.exists(os.path.join(path, 'info.xml')):
                paths.append(path)
        return paths

    def clearList(self):
        while self.pluginList.count() > 0:
            self.pluginList.takeItem(0)

    def showPlugins(self, search_str=None):
        from pkg_resources import parse_version
        self.clearList()
        if search_str == None or len(search_str) == 0:
            names = sorted(self.plugins.keys())
        else:
            def sort_func(name):
                name = str(name)
                return -difflib.SequenceMatcher(None, name.lower(), search_str.lower()).ratio() - int(search_str.lower() in name.lower())
            d = {name: sort_func(name) for name in self.plugins.keys() if sort_func(name) != 0}
            names = sorted(d.keys(), key=lambda a: d[a])
        for name in names:
            plug = PluginManager.plugins[name]
            if plug.version == '':
                plug.listWidget.setIcon(QtGui.QIcon())
            elif parse_version(plug.version) < parse_version(plug.latest_version):
                plug.listWidget.setIcon(QtGui.QIcon(image_path('exclamation.png')))
            else:
                plug.listWidget.setIcon(QtGui.QIcon(image_path('check.png')))
            self.pluginList.addItem(plug.listWidget)

    @staticmethod
    def removePlugin(plugin):
        PluginManager.gui.statusBar.showMessage("Uninstalling {}".format(plugin.name))
        if os.path.isdir(os.path.join(plugin_dir, plugin.directory, '.git')):
            g.alert("This plugin's directory is managed by git. To remove, manually delete the directory")
            return False
        try:
            shutil.rmtree(os.path.join(plugin_dir, plugin.directory), ignore_errors=True)
            plugin.version = ''
            plugin.menu = None
            plugin.listWidget.setIcon(QtGui.QIcon())
            PluginManager.gui.statusBar.showMessage('{} successfully uninstalled'.format(plugin.name))
        except Exception as e:
            g.alert(title="Plugin Uninstall Failed", msg="Unable to remove the folder at %s\n%s\nDelete the folder manually to uninstall the plugin" % (plugin.name, e), icon=QtWidgets.QMessageBox.Warning)

        PluginManager.gui.pluginSelected(plugin.listWidget)
        plugin.installed = False

    @staticmethod
    def downloadPlugin(plugin):
        PluginManager.gui.statusBar.showMessage("Installing plugin")
        import pip
        if isinstance(plugin, str):
            if plugin in PluginManager.plugins:
                plugin = PluginManager.plugins[plugin]
            else:
                return
        if plugin.url is None:
            return
        failed = []
        dists = [a.project_name for a in pkg_resources.working_set]
        PluginManager.gui.statusBar.showMessage("Installing dependencies for %s" % plugin.name)
        for pl in plugin.dependencies:
            try:
                if pl in dists:
                    continue
                a = __import__(pl)
            except ImportError:
                res = pip.main(['install', pl, '--no-cache-dir'])
                if res != 0:
                    failed.append(pl)
        if failed:
            if platform.system() == 'Windows':
                QtGui.QDesktopServices.openUrl(QtCore.QUrl("http://www.lfd.uci.edu/~gohlke/pythonlibs/#"+pl))
                v = str(sys.version_info.major) + str(sys.version_info.minor)
                if platform.architecture()[0]=='64bit':
                    arch = '_amd64'
                else:
                    arch = '32'
                g.alert("""Failed to install the dependency '{0}'. You must install {0} manually.
Download {0}-x-cp{1}-cp{1}m-win{2}.whl.

Once the wheel is downloaded, drag it into flika to install.

Then try installing the plugin again.""".format(pl, v, arch))
            else:
                g.alert("Failed to install dependencies for {}:\n{}\nYou must install them on your own before installing this plugin.".format(plugin.name, ', '.join(failed)))

            return

        if os.path.exists(os.path.join(plugin_dir, plugin.directory)):
            g.alert("A folder with name {} already exists in the plugins directory. Please remove it to install this plugin!".format(plugin.directory))
            return

        PluginManager.gui.statusBar.showMessage('Opening %s' % plugin.url)
        try:
            data = urlopen(plugin.url).read()
        except:
            g.alert(title="Download Error", msg="Failed to connect to %s to install the %s flika Plugin. Check your internet connection and try again, or download the plugin manually." % (PluginManager.gui.link, plugin.name), icon=QtWidgets.QMessageBox.Warning)
            return

        try:

            with tempfile.TemporaryFile() as tf:
                tf.write(data)
                tf.seek(0)
                with zipfile.ZipFile(tf) as z:
                    folder_name = os.path.dirname(z.namelist()[0])
                    z.extractall(plugin_dir)

            plugin = PluginManager.plugins[plugin.name]
            directory = os.path.join(plugin_dir, plugin.directory)
            os.rename(os.path.join(plugin_dir, folder_name), directory)
        except (PermissionError, Exception) as e:
            if os.path.exists(folder_name):
                shutil.rmtree(folder_name)
            if isinstance(e, PermissionError):
                g.alert("Unable to download plugin to {}. Rerun flika as administrator and download the plugin again.".format(plugin.name), title='Permission Denied')
            else:
                g.alert("Error occurred while installing {}.\n\t{}".format(plugin.name, e), title='Plugin Install Failed')    
            
            return
        
        PluginManager.gui.statusBar.showMessage('Extracting  %s' % plugin.name)
        plugin.version = plugin.latest_version
        plugin.listWidget.setIcon(QtGui.QIcon(image_path("check.png")))
        #plugin.menu = make_plugin_menu(plugin)
        plugin.menu = QtWidgets.QMenu(plugin.name)
        build_submenu(plugin.directory, plugin.menu, plugin.menu_layout)
        
        PluginManager.gui.statusBar.showMessage('Successfully installed {} and it\'s dependencies'.format(plugin.name))
        PluginManager.gui.pluginSelected(plugin.listWidget)
        plugin.installed = True


class Load_Local_Plugins_Thread(QtCore.QThread):
    plugins_done_sig = QtCore.Signal(dict)
    error_loading = QtCore.Signal(str)
    def __init__(self):
        QtCore.QThread.__init__(self)

    def __del__(self):
        self.wait()

    def run(self):
        #logger.debug("Started 'app.plugin_manager.load_local_plugins'")
        plugins = {n: Plugin(n) for n in plugin_list}
        installed_plugins = {}
        for pluginPath in PluginManager.local_plugin_paths():
            p = Plugin()
            p.fromLocal(pluginPath)
            try:
                p.bind_menu_and_methods()
                if p.name not in plugins.keys() or p.name not in installed_plugins.keys():
                    p.installed = True
                    plugins[p.name] = p
                    installed_plugins[p.name] = p
                else:
                    g.alert('Could not load the plugin {}. There is already a plugin with this same name. Change the plugin name in the info.xml file'.format(p.name))
            except Exception as e:
                msg = "Could not load plugin {}".format(pluginPath)
                self.error_loading.emit(msg)
                #g.alert(msg)
                logger.error(msg)
                ex_type, ex, tb = sys.exc_info()
                sys.excepthook(ex_type, ex, tb)
        self.plugins_done_sig.emit(plugins)
        #logger.debug("Completed 'app.plugin_manager.load_local_plugins'")

# from flika.app.plugin_manager import *


logger.debug("Completed 'reading app/plugin_manager.py'")


from .. import global_vars as g
from ..window import Window
import numpy as np
from ..roi import makeROI

class TestSettings():

	def test_random_roi_color(self):
		initial = g.settings['roi_color']
		g.settings['roi_color'] = 'random'
		w1 = Window(np.random.random([10, 10, 10]))
		roi1 = makeROI('rectangle', [[1, 1], [3, 3]])
		roi2 = makeROI('rectangle', [[2, 2], [3, 3]])
		assert roi1.pen.color().name() != roi2.pen.color().name(), 'Random ROI color is the same. This could be a random chance. Run repeatedly.'
		
		g.settings['roi_color'] = '#00ff00'
		roi3 = makeROI('rectangle', [[3, 3], [3, 3]])
		assert roi3.pen.color().name() == "#00ff00", 'ROI color set. all rois are same color'

		g.settings['roi_color'] = initial

	def test_multitrace(self):
		initial = g.settings['multipleTraceWindows']
		g.settings['multipleTraceWindows'] = False

		w1 = Window(np.random.random([10, 10, 10]))
		roi1 = makeROI('rectangle', [[1, 1], [3, 3]])
		roi1.plot()
		roi2 = makeROI('rectangle', [[2, 2], [3, 3]])
		roi2.plot()
		assert roi1.traceWindow == roi2.traceWindow, 'Traces not plotted together.'
		
		g.settings['multipleTraceWindows'] = True
		roi3 = makeROI('rectangle', [[3, 3], [3, 3]])
		roi3.plot()
		assert roi3.traceWindow != roi1.traceWindow, 'Multiple trace windows'

		g.settings['multipleTraceWindows'] = initial



import sys, os
import optparse

from ..process import *
from .. import global_vars as g
from ..window import Window
import numpy as np
import time
import pytest
from ..roi import makeROI, open_rois
import pyqtgraph as pg
from qtpy import QtGui

def test_example1():
	w = generate_random_image(120, 140)
	trimmed = trim(0, 0, keepSourceWindow=True)
	assert trimmed.imageview.ui.roiPlot.isVisible() == False
	assert trimmed.image.shape[-2:] == w.image.shape[-2:]
	roi = makeROI('rectangle', [[10, 10], [15, 5]], w)
	cropped = roi.crop()






from .logger import logger
logger.debug("Started 'reading __init__.py'")
from .version import __version__
from .flika import start_flika
from . import global_vars as g
logger.debug("Completed 'reading __init__.py'")


import sys, os
import optparse

from ..process.file_ import *
from .. import global_vars as g
from ..window import Window
import numpy as np
import time
import pytest
from ..roi import makeROI, open_rois
import pyqtgraph as pg
from qtpy import QtGui
im = np.random.random([150, 60, 60])



class Test_File():
	def test_open(self):
		if len(g.settings['recent_files']) == 0:
			g.settings['recent_files'].append('tests')
		w = open_file()
		w.close()

	def test_open_recent(self):
		g.m._make_recents()
		if len(g.m.recentFileMenu.actions()) > 0:
			g.m.recentFileMenu.actions()[0].trigger()
			g.win.close()

	def test_save_as(self):
		w = Window(im)
		a = save_file('movie')
		os.remove(a)
		w.close()

	def test_points_io(self):
		w = Window(im)
		w.addPoint([1, 2, 3])
		w.addPoint([3, 1, 2])
		ps = save_points('test.pos')
		open_points('test.pos')
		os.remove('test.pos')

		w.close()

	def test_rois_io(self):
		w = Window(im)
		a = makeROI('rectangle', [[3, 7], [6, 5]])
		w.save_rois('test.roi')
		a.delete()
		b = open_rois('test.roi')[0]
		os.remove('test.roi')
		assert np.array_equal(b.pts, [[3, 7], [6, 5]])
		w.close()
	



from logging import *
import os, sys


#LEVEL = DEBUG
LEVEL = WARNING


def get_log_file():
    LOG_DIR = os.path.join(os.path.expanduser("~"), '.FLIKA', 'log')
    MAX_LOG_IDX = 99
    if not os.path.exists(LOG_DIR):
        os.makedirs(LOG_DIR)
    existing_files = os.listdir(LOG_DIR)
    existing_files = [f for f in existing_files if os.path.splitext(f)[1] == '.log']
    if 'FLIKALOG.log' in existing_files: # This was the name of the log file before version 0.2.23
        try:
            os.remove(os.path.join(LOG_DIR, 'FLIKALOG.log'))
        except Exception:
            print("Cannot remove FLIKALOG.log file")
            existing_files.remove('FLIKALOG.log')
        else:
            existing_files = os.listdir(LOG_DIR)
            existing_files = [f for f in existing_files if os.path.splitext(f)[1] == '.log']

    try:
        existing_idxs = [int(f.split('.')[0]) for f in existing_files]
    except ValueError:
        print("There is an error with your log folder. Delete all files inside ~/.FLIKA/log/ and restart flika.")
    log_idx = 0
    while log_idx in existing_idxs:
        log_idx += 1
    log_idx = log_idx % MAX_LOG_IDX
    idx_to_delete = (int(log_idx) + 1) % MAX_LOG_IDX
    LOG_FILE_TO_DELETE = os.path.join(LOG_DIR, '{0:0>3}.log'.format(idx_to_delete))
    if os.path.exists(LOG_FILE_TO_DELETE):
        try:
            os.remove(LOG_FILE_TO_DELETE)
        except Exception as e:
            print(e)
    LOG_FILE = os.path.join(LOG_DIR, '{0:0>3}.log'.format(log_idx))
    return LOG_FILE

LOG_FILE = get_log_file()
FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
basicConfig(filename=LOG_FILE, format=FORMAT)
logger = getLogger("flika")
logger.setLevel(LEVEL)
handler = StreamHandler()
handler.setLevel(LEVEL)
formatter = Formatter(FORMAT)
handler.setFormatter(formatter)
logger.addHandler(handler)


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
sys.excepthook = handle_exception


# -*- coding: utf-8 -*-
from ..logger import logger
logger.debug("Started 'reading process/stacks.py'")


import numpy as np
from qtpy import QtWidgets
from ..window import Window
from .. import global_vars as g
from ..utils.BaseProcess import BaseProcess, BaseProcess_noPriorWindow, WindowSelector, MissingWindowError, CheckBox, SliderLabel, ComboBox



__all__ = ['deinterleave','trim','zproject','image_calculator', 'pixel_binning', 'frame_binning', 'resize', 'concatenate_stacks', 'duplicate', 'generate_random_image', 'change_datatype']

def duplicate():
    old = g.win
    if old is None:
        g.alert('Select a window before trying to duplicate it.')
    else:
        return Window(old.image, old.name, old.filename, old.commands, old.metadata)

class Generate_Random_Image(BaseProcess_noPriorWindow):
    def __init__(self):
        super().__init__()
        self.puffs=None

    def get_init_settings_dict(self):
        s = dict()
        s['nFrames'] = 1000
        s['width'] = 128
        return s

    def gui(self):
        self.gui_reset()
        nFrames = SliderLabel(0)
        nFrames.setRange(0,10000)
        width = SliderLabel(0)
        width.setRange(2,1024)
        self.items.append({'name':'nFrames','string':'Movie Duration (frames)','object':nFrames})
        self.items.append({'name':'width','string':'Width of image (pixels)','object':width})
        super().gui()

    def __call__(self, nFrames=10000, width=128):
        self.start()
        self.newtif = np.random.randn(nFrames, width, width)
        self.newname=' Random Noise '
        return self.end()
generate_random_image = Generate_Random_Image()

class Deinterleave(BaseProcess):
    """deinterleave(nChannels, keepSourceWindow=False)

    This deinterleaves a stack into nChannels
    
    Parameters:
        nChannels (int): The number of channels to deinterleave.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        return {'nChannels': 3}

    def gui(self):
        self.gui_reset()
        nChannels=QtWidgets.QSpinBox()
        nChannels.setMinimum(2)
        self.items.append({'name':'nChannels','string':'How many Channels?','object':nChannels})
        super().gui()
    def __call__(self,nChannels,keepSourceWindow=False):
        self.start(keepSourceWindow)

        newWindows=[]
        for i in np.arange(nChannels):
            newtif=self.tif[i::nChannels]
            name=self.oldname+' - Channel '+str(i)
            newWindow=Window(newtif,name,self.oldwindow.filename)
            newWindows.append(newWindow)
        
        if keepSourceWindow is False:
            self.oldwindow.close()  
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        return newWindows
deinterleave=Deinterleave()

class Pixel_binning(BaseProcess):
    """pixel_binning(nPixels, keepSourceWindow=False)

    This bins the pixels to reduce the file size
    
    Parameters:
        nPixels (int): The number of pixels to bin.  Example: a value of 2 will reduce file size from 256x256->128x128.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        s = dict()
        s['nPixels'] = 2
        return s

    def gui(self):
        self.gui_reset()
        nPixels=QtWidgets.QSpinBox()
        nPixels.setMinimum(2)
        nPixels.setMaximum(2)
        self.items.append({'name': 'nPixels', 'string': 'How many adjacent pixels to bin?', 'object': nPixels})
        super().gui()
    def __call__(self,nPixels,keepSourceWindow=False):
        self.start(keepSourceWindow)
        tif=self.tif
        nDim=len(tif.shape)
        if nPixels==2:
            if nDim==4:
                image1=tif[:,0::2,0::2,:]
                image2=tif[:,1::2,0::2,:]
                image3=tif[:,0::2,1::2,:]
                image4=tif[:,1::2,1::2,:]
                self.newtif=(image1+image2+image3+image4)/4
                self.newname=self.oldname+' - Binned'
                return self.end()
            elif nDim==3:
                mt,mx,my=tif.shape
                if mx%2==1:
                    tif=tif[:,:-1,:]
                if my%2==1:
                    tif=tif[:,:,:-1]
                image1=tif[:,0::2,0::2]
                image2=tif[:,1::2,0::2]
                image3=tif[:,0::2,1::2]
                image4=tif[:,1::2,1::2]
                self.newtif=(image1+image2+image3+image4)/4
                self.newname=self.oldname+' - Binned'
                return self.end()
            elif nDim==2:
                image1=tif[0::2,0::2]
                image2=tif[1::2,0::2]
                image3=tif[0::2,1::2]
                image4=tif[1::2,1::2]
                self.newtif=(image1+image2+image3+image4)/4
                self.newname=self.oldname+' - Binned'
                return self.end()
        else:
            g.alert("2 is the only supported value for binning at the moment")
pixel_binning=Pixel_binning()

class Frame_binning(BaseProcess):
    """frame_binning(nFrames, keepSourceWindow=False)

    This bins the pixels to reduce the file size
    
    Parameters:
        nFrames (int): The number of frames to bin.  Example: a value of 2 will reduce number of frames from 1000 to 500.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        return {'nFrames': 2}

    def gui(self):
        self.gui_reset()
        nFrames=QtWidgets.QSpinBox()
        nFrames.setMinimum(2)
        nFrames.setMaximum(10000)
        self.items.append({'name':'nFrames','string':'How many frames to bin?','object':nFrames})
        super().gui()
    def __call__(self,nFrames,keepSourceWindow=False):
        self.start(keepSourceWindow)
        tif=self.tif
        nDim=len(tif.shape)
        self.newtif=np.array([np.mean(tif[i:i+nFrames],0) for i in np.arange(0,len(tif),nFrames)])
        self.newname=self.oldname+' - Binned {} frames'.format(nFrames)
        return self.end()
frame_binning=Frame_binning()



class Resize(BaseProcess):
    """resize(factor, keepSourceWindow=False)

    Performs interpolation to up-size images
    
    Parameters:
        factor (int): The factor to scale the images by.  Example: a value of 2 will double the number of pixels wide the images are.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        return {'factor': 2}

    def gui(self):
        self.gui_reset()
        factor=QtWidgets.QSpinBox()
        factor.setMinimum(2)
        factor.setMaximum(100)
        self.items.append({'name':'factor','string':'By what factor to resize the image?','object':factor})
        super().gui()

    def __call__(self,factor,keepSourceWindow=False):
        import skimage.transform
        self.start(keepSourceWindow)
        if self.tif.dtype in (np.uint64, np.int64):
            g.alert("Resize fails on int64 and uint64 movie types, change the image type to resize.")
            return
        A = self.tif
        nDim = len(A.shape)
        is_rgb = self.oldwindow.metadata['is_rgb'] or (nDim == 3 and A.shape[2] == 3) or nDim == 4
        B = None
        if not is_rgb:
            if nDim == 3:
                mt,mx,my = A.shape
                B = np.zeros((mt, mx*factor, my*factor))
                for t in np.arange(mt):
                    B[t]=skimage.transform.resize(A[t], (mx*factor,my*factor))
            elif nDim == 2:
                mx, my = A.shape
                B = skimage.transform.resize(A,(mx*factor, my*factor))
        self.newtif = B
        self.newname = self.oldname+' - resized {}x '.format(factor)
        return self.end()
resize = Resize()


class Trim(BaseProcess):
    """trim(firstFrame, lastFrame, increment=1, delete=False, keepSourceWindow=False)

    This creates a new stack from the frames between the firstFrame and the lastFrame
    
    Parameters:
        firstFrame (int): The index of the first frame in the stack to be kept.
        lastFrame (int): The index of the last frame in the stack to be kept.
        increment (int): if increment equals i, then every ith frame is kept.
        delete (bool): if False, then the specified frames will be kept.  If True, they will be deleted.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        s = dict()
        s['firstFrame'] = 0
        s['lastFrame'] = g.win.image.shape[0]
        s['increment'] = 1
        s['delete'] = False
        return s

    def gui(self):
        self.gui_reset()
        nFrames=1
        if g.win is not None:
            nFrames=g.win.image.shape[0]
        firstFrame=QtWidgets.QSpinBox()
        firstFrame.setMaximum(nFrames-1)
        lastFrame=QtWidgets.QSpinBox()
        lastFrame.setRange(0,nFrames-1)
        increment=QtWidgets.QSpinBox()
        increment.setMaximum(nFrames)
        increment.setMinimum(1)
        delete = CheckBox()

        self.items.append({'name': 'firstFrame', 'string': 'First Frame', 'object': firstFrame})
        self.items.append({'name': 'lastFrame',  'string': 'Last Frame',  'object': lastFrame })
        self.items.append({'name': 'increment',  'string': 'Increment',   'object': increment })
        self.items.append({'name': 'delete',     'string': 'Delete',      'object': delete    })
        super().gui()

    def __call__(self, firstFrame, lastFrame, increment=1, delete = False, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if not delete:
            self.newtif=self.tif[firstFrame:lastFrame+1:increment]
        if delete:
            idxs_not=np.arange(firstFrame, lastFrame+1, increment)
            idxs = np.ones(len(self.tif),dtype=np.bool)
            idxs[idxs_not] = False
            self.newtif = self.tif[idxs]
        self.newname=self.oldname+' - Kept Stack'
        return self.end()
trim=Trim()


class ZProject(BaseProcess):
    """zproject(firstFrame, lastFrame, projection_type, keepSourceWindow=False)

    This creates a new stack from the frames between the firstFrame and the lastFrame
    
    Parameters:
        firstFrame (int): The index of the first frame in the stack to be kept.
        lastFrame (int): The index of the last frame in the stack to be kept.
        projection_type (str): Method used to combine the frames.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def get_init_settings_dict(self):
        return {'firstFrame': 0, 'lastFrame': 0, 'projection_type': 'Average'}

    def gui(self):
        self.gui_reset()
        nFrames=1
        if g.win and len(g.win.image.shape) != 3:
            g.alert('zproject only works on 3 dimensional windows')
            return False
        if g.win is not None:
            nFrames=g.win.image.shape[0]
        firstFrame=QtWidgets.QSpinBox()
        firstFrame.setMaximum(nFrames)
        self.items.append({'name':'firstFrame','string':'First Frame','object':firstFrame})
        lastFrame=QtWidgets.QSpinBox()
        lastFrame.setRange(1,nFrames-1)
        self.items.append({'name':'lastFrame','string':'Last Frame','object':lastFrame})
        projection_type=ComboBox()
        projection_type.addItem('Average')
        projection_type.addItem('Max Intensity')
        projection_type.addItem('Min Intensity')
        projection_type.addItem('Sum Slices')
        projection_type.addItem('Standard Deviation')
        projection_type.addItem('Median')
        self.items.append({'name':'projection_type','string':'Projection Type','object':projection_type})
        super().gui()
        lastFrame.setValue(nFrames - 1)

    def __call__(self, firstFrame, lastFrame, projection_type, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.ndim != 3 or self.tif.shape[2] == 3:
            g.m.statusBar().showMessage('zproject only works on 3 dimensional, non-color windows')
            return False
        self.newtif=self.tif[firstFrame:lastFrame+1]
        p=projection_type
        if p=='Average':
            self.newtif=np.mean(self.newtif,0)
        elif p=='Max Intensity':
            self.newtif=np.max(self.newtif,0)
        elif p=='Min Intensity':
            self.newtif=np.min(self.newtif,0)
        elif p=='Sum Slices':
            self.newtif=np.sum(self.newtif,0)
        elif p=='Standard Deviation':
            self.newtif=np.std(self.newtif,0)
        elif p=='Median':
            self.newtif=np.median(self.newtif,0)
        self.newname=self.oldname+' {} Projection'.format(projection_type)
        return self.end()
zproject=ZProject()

class Image_calculator(BaseProcess):
    """###image_calculator(window1,window2,operation,keepSourceWindow=False)

    This creates a new stack by combining two windows in an operation
    
    Parameters:
        window1 (Window): The first window
        window2 (Window): The second window
        operation (str): Method used to combine the frames.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        window1=WindowSelector()
        window2=WindowSelector()      
        operation=ComboBox()
        operation.addItems(['Add','Subtract','Multiply','Divide','AND','OR','XOR','Min','Max','Average'])
        self.items.append({'name':'window1','string':'Window 1','object':window1})
        self.items.append({'name':'window2','string':'Window 2','object':window2})
        self.items.append({'name':'operation','string':'Operation','object':operation})
        super().gui()
    def __call__(self,window1, window2, operation, keepSourceWindow=False):
        self.keepSourceWindow=keepSourceWindow
        g.m.statusBar().showMessage('Performing {}...'.format(self.__name__))
        if window1 is None or window2 is None:
            g.alert("You cannot execute '{}' without selecting a window first.".format(self.__name__))
            return None
        A = window1.image
        B = window2.image
        nDim1 = len(A.shape)
        nDim2 = len(B.shape)
        xyshape1 = A.shape
        xyshape2 = B.shape
        if nDim1 == 3:
            xyshape1 = xyshape1[1:]
        if nDim2 == 3:
            xyshape2 = xyshape2[1:]
        if xyshape1 != xyshape2:
            g.alert('The two windows have images of different shapes. They could not be combined')
            return None
        if nDim1 == 3 and nDim2 == 3:
            n1=A.shape[0]; n2=B.shape[0]
            if n1 != n2: #if the two movies have different # frames
                n = np.min([n1,n2])
                A = A[:n] #shrink them so they have the same length
                B = B[:n]
                
        
        if operation=='Add':
            self.newtif=np.add(A,B)
        elif operation=='Subtract':
            self.newtif=np.subtract(A,B)
        elif operation=='Multiply':
            self.newtif=np.multiply(A,B)
        elif operation=='Divide':
            self.newtif=np.divide(A,B)
            self.newtif[np.isnan(self.newtif)]=0
            self.newtif[np.isinf(self.newtif)]=0
        if operation=='AND':
            self.newtif=np.logical_and(window1.image,window2.image).astype(np.uint8)
        elif operation=='OR':
            self.newtif=np.logical_or(window1.image,window2.image).astype(np.uint8)
        elif operation=='XOR':
            self.newtif=np.logical_xor(window1.image,window2.image).astype(np.uint8)
        elif operation=='Min':
            self.newtif=np.minimum(A,B)
        elif operation=='Max':
            self.newtif=np.maximum(A,B)
        elif operation=='Average':
            if nDim1==3 and nDim2==3:
                C=np.concatenate((np.expand_dims(A,4),np.expand_dims(B,4)),3)
                self.newtif=np.mean(C,3)
            elif nDim1==2 and nDim2==2:
                C=np.concatenate((np.expand_dims(A,3),np.expand_dims(B,3)),2)
                self.newtif=np.mean(C,2)
            else:
                if nDim1==3:
                    B=np.repeat(np.expand_dims(B,0),len(A),0)
                elif nDim2==3:
                    A=np.repeat(np.expand_dims(A,0),len(B),0)
                C=np.concatenate((np.expand_dims(A,4),np.expand_dims(B,4)),3)
                self.newtif=np.mean(C,3)
#            
        self.oldwindow=window1
        self.oldname=window1.name
        self.newname=self.oldname+' - {}'.format(operation)
        if keepSourceWindow is False:
            print('closing both windows')
            window1.close()
            window2.close()
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        newWindow=Window(self.newtif,str(self.newname),self.oldwindow.filename)
        del self.newtif
        return newWindow
        
image_calculator=Image_calculator()


class Concatenate_stacks(BaseProcess):
    """concatenate_stacks(window1,window2,keepSourceWindow=False)

    This creates a new stack by concatenating two stacks

    Parameters:
        window1 (Window): The first window
        window2 (Window): The second window
    Returns:
        newWindow
    """

    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        window1 = WindowSelector()
        window2 = WindowSelector()
        self.items.append({'name': 'window1', 'string': 'Window 1', 'object': window1})
        self.items.append({'name': 'window2', 'string': 'Window 2', 'object': window2})
        super().gui()

    def __call__(self, window1, window2, keepSourceWindow=False):
        self.keepSourceWindow = keepSourceWindow
        g.m.statusBar().showMessage('Performing {}...'.format(self.__name__))
        if window1 is None or window2 is None:
            raise (
            MissingWindowError("You cannot execute '{}' without selecting a window first.".format(self.__name__)))
        A = window1.image
        B = window2.image
        try:
            self.newtif = np.concatenate((A, B), 0)
        except ValueError:
            if len(A.shape) == 2 and len(B.shape) == 3:
                self.newtif = np.concatenate((np.expand_dims(A, 0), B), 0)
            elif len(A.shape) == 3 and len(B.shape) == 2:
                self.newtif = np.concatenate((A, np.expand_dims(B, 0)), 0)
        self.oldwindow = window1
        self.oldname = window1.name
        self.newname = self.oldname + ' - {}'.format('concatenated')
        if keepSourceWindow is False:
            print('closing both windows')
            window1.close()
            window2.close()
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        newWindow = Window(self.newtif, str(self.newname), self.oldwindow.filename)
        del self.newtif
        return newWindow


concatenate_stacks = Concatenate_stacks()

class Change_datatype(BaseProcess):
    """change_datatype(datatype, keepSourceWindow=False)

    This bins the pixels to reduce the file size

    Parameters:
        datatype (string)
    Returns:
        newWindow
    """

    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        dtype = ComboBox()
        dtype_strs = ['uint8', 'uint16', 'uint32', 'uint64', 'int8','int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        for d in dtype_strs:
            dtype.addItem(d)
        self.items.append({'name': 'datatype', 'string': 'Convert image to which datatype?', 'object': dtype})
        super().gui()
    def __call__(self, datatype, keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif = self.tif.astype(np.dtype(datatype))
        self.newname = self.oldname
        return self.end()

change_datatype = Change_datatype()

logger.debug("Completed 'reading process/stacks.py'")



# -*- coding: utf-8 -*-
import os, sys
from urllib.request import urlopen
from urllib.error import HTTPError
import re
import sys
from io import StringIO, BytesIO
import contextlib
import pathlib
import tempfile
from zipfile import ZipFile
import shutil
from qtpy import QtWidgets, QtGui, QtCore
from . import global_vars as g
from .version import __version__ as installed_flika_version



__all__ = ['checkUpdates']


def check_if_installed_via_pip():
    from distutils.sysconfig import get_python_lib
    s_loc = pathlib.Path(get_python_lib())
    f_loc = pathlib.Path(__file__)
    assert f_loc.exists()
    assert s_loc.exists()
    return s_loc in f_loc.parents


@contextlib.contextmanager
def capture():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


def path_walk(top, topdown=False, followlinks=False):
    """
         See Python docs for os.walk, exact same behavior but it yields Path() instances instead
    """
    names = list(top.iterdir())

    dirs = (node for node in names if node.is_dir() is True)
    nondirs = (node for node in names if node.is_dir() is False)

    if topdown:
        yield top, dirs, nondirs

    for name in dirs:
        if followlinks or name.is_symlink() is False:
            for x in path_walk(name, topdown, followlinks):
                yield x

    if topdown is not True:
        yield top, dirs, nondirs


def get_pypi_version():
    import pip
    with capture() as out:
        pip.main(['search', 'flika'])
    stdout, stderr = out
    pypi_version = re.search(r'\((.*?)\)', stdout).group(1)
    return pypi_version


def get_github_version():
    url = "https://raw.githubusercontent.com/flika-org/flika/master/flika/version.py"
    try:
        data = urlopen(url).read().decode('utf-8')
    except Exception as e:
        g.alert("Connection Failed",
                "Cannot connect to flika Repository. Connect to the internet to check for updates. %s" % e)
        return
    github_version = re.match(r'__version__\s*=\s*\'([\d\.\']*)\'', data).group(1)
    return github_version


def checkUpdates():
    """
    If flika was installed via pip, this function uses pip to lookup the current version.

    If flika was downloaded directly from Github, this will check the current version in the master branch of the Github
    repository.

    """
    from pkg_resources import parse_version
    installed_via_pip = check_if_installed_via_pip()
    if installed_via_pip:
        latest_version = get_pypi_version()
    else:
        latest_version = get_github_version()
    message = "Installed version: " + installed_flika_version
    if latest_version is None:
        latest_version = 'Unknown'
    message += '\nLatest Version: ' + latest_version

    if parse_version(installed_flika_version) < parse_version(latest_version):
        if g.messageBox("Update Recommended", message + '\n\nWould you like to update?',
                      QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                      QtWidgets.QMessageBox.Question) == QtWidgets.QMessageBox.Yes:
            updateFlika()
    else:
        g.messageBox("Up to date", "Your version of Flika is up to date\n" + message)


def updateFlika():
    import pip
    installed_via_pip = check_if_installed_via_pip()
    if installed_via_pip:
        with capture() as out:
            pip.main(['install', '-U', '--no-deps', 'flika'])
        stdout, stderr = out
        print(stdout)
        g.alert('Update successful. Restart flika to complete update.')
    else:
        flika_location = pathlib.Path(__file__).parents[1]
        assert flika_location.stem == 'flika'
        if flika_location.joinpath('.git').exists():
            g.alert("This installation of flika is managed by git. Upgrade will not proceed. Use git to upgrade.")
            return False
        extract_location = tempfile.mkdtemp()
        url = urlopen('https://github.com/flika-org/flika/archive/master.zip')
        print("Downloading flika from Github to {}".format(extract_location))
        try:
            with ZipFile(BytesIO(url.read())) as z:
                folder_name = os.path.dirname(z.namelist()[0])
                if os.path.exists(extract_location):
                    shutil.rmtree(extract_location)
                z.extractall(extract_location)
        except Exception as e:
            g.messageBox("Update Error", "Failed to remove and replace old version of flika. %s" % e,
                         icon=QtWidgets.QMessageBox.Warning)
            return False
        extract_location = pathlib.Path(extract_location)
        new_flika_location = extract_location.joinpath('flika-master')
        assert new_flika_location.exists()
        try:
            for src_path, subs, fs in path_walk(new_flika_location):
                dst_path = flika_location.joinpath(src_path.relative_to(new_flika_location))
                for src_f in fs:
                    if src_f.__str__().endswith(('.py', '.ui', '.png', '.txt', '.xml', '.in', '.ico', '.rst')):
                        g.m.statusBar().showMessage('replacing %s' % src_f)
                        dst_f = flika_location.joinpath(src_f.relative_to(new_flika_location))
                        shutil.copy(src_f.__str__(), dst_f.__str__())
        except PermissionError as e:
            g.messageBox("Update Error", "Failed to remove and replace old version of flika. %s" % e,
                         icon=QtWidgets.QMessageBox.Warning)
            shutil.rmtree(extract_location.__str__())
            return False
        shutil.rmtree(extract_location.__str__())
        g.alert('Update successful. Restart flika to complete update.')


# -*- coding: utf-8 -*-
from .logger import logger
logger.debug("Started 'reading flika.py'")
import sys, os
import platform
import optparse
import warnings
logger.debug("Started 'reading flika.py, importing numpy'")
import numpy as np
logger.debug("Completed 'reading flika.py, importing numpy'")
from .version import __version__
from .app.application import FlikaApplication


# for development purposes, add this if flika is not in your site-packages
# sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))



warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

def parse_arguments(argv):
    ''' Parses command line arguments for valid flika args

    Arguments:
        argv: Arguments passed to program

    Returns:
        A tuple of options, position arguments
    '''
    usage = """usage: %prog [FILE FILE...]

    # start a new session
    %prog

    # start a new session and load a file
    %prog image.tiff

    #start a new session with multiple files
    %prog image.tiff script.py

    #run a script
    %prog -x script.py

    #increase verbosity level
    %prog -v

    #run the test suite0

    %prog -t
    """
    parser = optparse.OptionParser(usage=usage, version=str(__version__))

    parser.add_option('-x', '--execute', action='store_true', dest='script',
                      help="Open file in script editor and run", default=False)
    parser.add_option('-t', '--test', action='store_true', dest='test',
                      help="Run test suite", default=False)
    parser.add_option('-v', '--verbose', action='store_true',
                      help="Increase the vebosity level", default=False)

    err_msg = verify(parser, argv)
    if err_msg:
        sys.stderr.write('\n%s\n' % err_msg)
        parser.print_help()
        sys.exit(1)

    return parser.parse_args(argv)

def verify(parser, argv):
    """verify(parser, argv)
    Check for input errors

    Arguments:
        parser: OptionParser instance
        argv (list): Argument list

    Returns:
        An error message in the event of an input error, or None
    """
    opts, args = parser.parse_args(argv)
    err_msg = None
    if opts.script and len(args) != 1:
        err_msg = "Must provide a script\n"

    return err_msg

def ipython_qt_event_loop_setup():
    try:
        __IPYTHON__
    except NameError:
        return #  If __IPYTHON__ is not defined, we are not in ipython
    else:
        print("Starting flika inside IPython")
        from IPython import get_ipython
        ipython = get_ipython()
        ipython.magic("gui qt")

def load_files(files):
    from .process.file_ import open_file
    for f in files:
        open_file(f)

def start_flika(files=[]):
    """Run a flika session and exit, beginning the event loop

    Parameters:
        files (list): An optional list of data files to load.

    Returns:
        A flika application object with optional files loaded
 
    """
    logger.debug("Started 'flika.start_flika()'")
    print('Starting flika')
    fa = FlikaApplication()
    load_files(files)
    fa.start()
    ipython_qt_event_loop_setup()
    logger.debug("Completed 'flika.start_flika()'")
    return fa

def exec_():
    fa = start_flika(sys.argv[1:])
    return fa.app.exec_()

def post_install():
    if platform.system() == 'Windows':
        print("Creating start menu shortcut...")
        import winshell
        icon_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'images', 'favicon.ico')
        flika_exe = os.path.join(sys.exec_prefix, 'Scripts', 'flika.exe')
        link_path = os.path.join(winshell.programs(), "flika.lnk")
        with winshell.shortcut(link_path) as link:
            link.path = flika_exe
            link.description = "flika"
            link.icon_location = (icon_path, 0)
        link_path = os.path.join(winshell.desktop(), "flika.lnk")
        with winshell.shortcut(link_path) as link:
            link.path = flika_exe
            link.description = "flika"
            link.icon_location = (icon_path, 0)

if __name__ == '__main__':
    start_flika(sys.argv[1:])

logger.debug("Completed 'reading flika.py'")
"""
def exec_(args=sys.argv):
    opt, args = parse_arguments(args[1:])

    if opt.verbose:
        logger.setLevel("INFO")

    logger.info("Input arguments: %s", sys.argv)

    start_flika(files=args)


def run(args=sys.argv):
    ''' open flika without running exec_. For debugging purposes
    '''
    opt, args = parse_arguments(args[1:])

    if opt.verbose:
        logger.setLevel("INFO")

    fa = FlikaApplication()
    fa.show()

    load_files(files=args)

    if 'PYCHARM_HOSTED' not in os.environ and 'SPYDER_SHELL_ID' not in os.environ:
        return fa.app.exec_()
"""




# -*- coding: utf-8 -*-
from ..logger import logger
logger.debug("Started 'reading app/script_editor.py'")
from qtpy import QtGui, QtCore, QtWidgets, uic
import os

from .. import global_vars as g
from .syntax import PythonHighlighter
from ..utils.misc import save_file_gui, open_file_gui, load_ui

MESSAGE_TIME = 2000
try:
    __IPYTHON__
except NameError:
    INSIDE_IPYTHON = False
else:
    INSIDE_IPYTHON = True

def qstr2str(string):
    string=str(string)
    return string.replace(u'\u2029','\n').strip()

class Editor(QtWidgets.QPlainTextEdit):
    def __init__(self, scriptfile = ''):
        super(Editor, self).__init__()
        self.highlight = PythonHighlighter(self.document())
        self.scriptfile = ''
        if scriptfile != '':
            self.open_file(scriptfile)
        font = QtGui.QFont()
        font.setFamily("Courier")
        font.setStyleHint(QtGui.QFont.Monospace)
        font.setFixedPitch(True)
        font.setPointSize(ScriptEditor.POINT_SIZE)

        self.setFont(font)

        metrics = QtGui.QFontMetrics(font)
        self.setTabStopWidth(4 * metrics.width(' '))

    @staticmethod
    def fromWindow(window):
        if window is None:
            g.alert("In order to load a script from a window, you need to have a window selected.")
            return None
        editor = Editor()
        editor.setPlainText('\n'.join(window.commands))
        return editor
        
    def open_file(self, scriptfile):
        self.scriptfile = scriptfile
        try:
            script = open(scriptfile, 'r').read()
        except FileNotFoundError as e:
            print("Failed to read %s: %s" % (scriptfile, e))
            return
        self.setPlainText(script)
        ScriptEditor.gui.statusBar().showMessage('{} opened.'.format(os.path.basename(self.scriptfile)), MESSAGE_TIME)

    def save_as(self):
        filename = save_file_gui('Save script', ScriptEditor.most_recent_script(), '*.py')
        if filename == '':
            ScriptEditor.gui.statusBar().showMessage('Save cancelled',  MESSAGE_TIME)
            return False
        self.scriptfile = filename
        self.save()
        return True

    def save(self):
        if not self.scriptfile:
            return self.save_as()
        f = open(self.scriptfile, 'w')
        command=qstr2str(self.toPlainText())
        f.write(command)
        f.close()
        ScriptEditor.add_recent_file(self.scriptfile)
        ScriptEditor.gui.statusBar().showMessage('{} saved.'.format(os.path.basename(self.scriptfile)), MESSAGE_TIME)
        return True

class ScriptEditor(QtWidgets.QMainWindow):
    '''
    QMainWindow for editing and running user scripts. Comprised of a tabbed text editor and console.
    '''
    POINT_SIZE = 8
    def __init__(self, parent=None, ):
        from .terminal import ipython_terminal
        from .script_namespace import getnamespace
        super(ScriptEditor, self).__init__(parent)
        load_ui('ipythonWidget.ui', self, directory=os.path.dirname(__file__))
        text = """Predefined Libraries:
    scipy and numpy (np)
Useful variables:
    g.m.windows - list of windows.
    g.win - the selected window
    g.win.rois -list of rois in that window
    - roi.getTrace() gets an roi trace as an array
    clear() to clear the console 
    reset() to reset all global and local variables """

        self.terminal = ipython_terminal(banner=text, **getnamespace())
        self.terminal.namespace.update({'clear': self.terminal.clear, 'reset': self.resetNamespace})
        layout = QtWidgets.QGridLayout()
        self.terminalWidget.setLayout(layout)
        layout.addWidget(self.terminal)
        self.addEditor()
        self.saveButton.clicked.connect(self.saveCurrentScript)
        self.runButton.clicked.connect(self.runScript)
        self.runSelectedButton.clicked.connect(self.runSelected)
        self.actionNew_Script.triggered.connect(lambda f: self.addEditor())
        self.actionFrom_File.triggered.connect(lambda f: ScriptEditor.importScript())
        self.actionFrom_Window.triggered.connect(lambda : self.addEditor(Editor.fromWindow(g.win)))
        self.actionSave_Script.triggered.connect(self.saveCurrentScript)
        self.menuRecentScripts.aboutToShow.connect(self.open_scripts)
        self.actionChangeFontSize.triggered.connect(self.changeFontSize)
        #self.eventeater = ScriptEventEater(self)
        self.setAcceptDrops(True)
        #self.installEventFilter(self.eventeater)
        self.scriptTabs.tabCloseRequested.connect(self.closeTab)
        self.setWindowTitle('Script Editor')

        
    def resetNamespace(self):
        self.terminal.shell.reset()
        self.terminal.shell.user_ns.update(getnamespace())
        self.terminal.namespace.update({'clear': self.terminal.clear, 'reset': self.resetNamespace})

    def changeFontSize(self):
        val, ok = QtWidgets.QInputDialog.getInt(self, "Change Font Size", "Font Size", \
            value=ScriptEditor.POINT_SIZE, min = 6, max = 50, step = 1)
        if ok:
            self.setFontSize(val)

    def setFontSize(self, val):
        ScriptEditor.POINT_SIZE = val

        font = self.scriptTabs.font()
        font.setPointSize(val)
        self.scriptTabs.setFont(font)

        f = self.terminal._control.font()
        f.setPointSize(val)
        self.terminal._control.setFont(f)

        for ch in self.scriptTabs.children():
            if isinstance(ch, QtWidgets.QStackedWidget):
                for i in range(ch.count()):
                    f = ch.widget(i).font()
                    f.setPointSize(val)
                    ch.widget(i).setFont(f)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        for url in event.mimeData().urls():
            filename=url.toString()
            filename=filename.split('file:///')[1]
            print('filename={}'.format(filename))
            ScriptEditor.importScript(filename)

        event.accept()

    def open_scripts(self):
        self.menuRecentScripts.clear()
        def makeFun(script):
            return lambda: ScriptEditor.importScript(script)
        if len(g.settings['recent_scripts']) == 0:
            action = QtWidgets.QAction("No Recent Scripts", self)
            action.setEnabled(False)
            self.menuRecentScripts.addAction(action)
        else:
            for filename in g.settings['recent_scripts']:
                if not os.path.exists(filename):
                    g.settings['recent_scripts'].remove(filename)
                    continue
                action = QtWidgets.QAction(filename, self, triggered=makeFun(filename))
                self.menuRecentScripts.addAction(action)

    def closeTab(self, index):
        self.scriptTabs.removeTab(index)

    def saveCurrentScript(self):
        cw = self.currentTab()
        if cw == None:
            return
        if cw.save():
            self.scriptTabs.setTabText(self.scriptTabs.currentIndex(), os.path.basename(cw.scriptfile))

    def currentTab(self):
        return self.scriptTabs.currentWidget()

    @staticmethod
    def most_recent_script():
        return g.settings['recent_scripts'][-1] if len(g.settings['recent_scripts']) > 0 else ''

    @staticmethod
    def add_recent_file(filename):
        filename = os.path.abspath(filename)
        if not os.path.exists(filename):
            return
        if filename in g.settings['recent_scripts']:
            g.settings['recent_scripts'].remove(filename)
        g.settings['recent_scripts'].insert(0, filename)
        if len(g.settings['recent_scripts']) > 8:
            g.settings['recent_scripts'] = g.settings['recent_scripts'][:-1]
        g.settings.save()

    @staticmethod
    def importScript(scriptfile = ''):
        if not hasattr(ScriptEditor, 'gui') or not ScriptEditor.gui.isVisible():
            ScriptEditor.show()
        if scriptfile == '':
            prompt = "Open script"
            directory = os.path.dirname(ScriptEditor.most_recent_script())
            filetypes = '*.py'
            scriptfile = open_file_gui(prompt, directory, filetypes)
            if scriptfile is None:
                return None
        if hasattr(ScriptEditor, 'gui'):
            editor = Editor(scriptfile)
            ScriptEditor.add_recent_file(scriptfile)
            ScriptEditor.gui.addEditor(editor)
    
    def addEditor(self, editor=None):
        self.setUpdatesEnabled(False)
        if editor is None:
            editor = Editor()
        if editor.scriptfile == '':
            name = 'New Script'
        else:
            name = editor.scriptfile
        editor.keyPressEvent = lambda ev: self.editorKeyPressEvent(editor, ev)
        self.scriptTabs.insertTab(0, editor, os.path.basename(name))
        self.scriptTabs.setCurrentIndex(0)
        self.setUpdatesEnabled(True)

        return editor

    def keyPressEvent(self, ev):
        if ev.modifiers() == QtCore.Qt.ControlModifier:
            if ev.key() == QtCore.Qt.Key_N:
                self.addEditor()
                ev.accept()
            elif ev.key() == QtCore.Qt.Key_S:
                self.saveCurrentScript()

    def editorKeyPressEvent(self, editor, ev):
        if ev.key() == QtCore.Qt.Key_F9:
            self.runSelected()
            ev.accept()
        else:
            Editor.keyPressEvent(editor, ev)

    def runScript(self):
        if self.currentTab() == None:
            return
        command=qstr2str(self.currentTab().toPlainText())
        if command:
            self.terminal.execute(command)

    def runSelected(self):
        if self.currentTab() == None:
            return
        cursor=self.currentTab().textCursor()
        command=cursor.selectedText()
        self.command=command
        command=qstr2str(command)
        if command:
            self.terminal.execute(command)

    @staticmethod
    def show():
        if 'PYCHARM_HOSTED' in os.environ:
            g.alert('You cannot run the script editor from within PyCharm.')
        elif INSIDE_IPYTHON:
            g.alert('You cannot run the script editor because flika is already running inside IPython.')
        else:
            if not hasattr(ScriptEditor, 'gui'):
                ScriptEditor.gui = ScriptEditor()
            elif ScriptEditor.gui.isVisible():
                return
            QtWidgets.QMainWindow.show(ScriptEditor.gui)

            ui = ScriptEditor.gui
            geom = ui.geometry()
            geom = (geom.x(), geom.y(), geom.width(), geom.height())
            settings = {'geometry': geom, 'sizes': ui.splitter.sizes()}
            if 'script_editor_settings' in g.settings:
                settings.update(g.settings['script_editor_settings'])
            ui.setGeometry(*settings['geometry'])
            ui.splitter.setSizes(settings['sizes'])

    def closeEvent(self, ev):
        geom = self.geometry()
        geom = (geom.x(), geom.y(), geom.width(), geom.height())
        g.settings['script_editor_settings'] = {'geometry': geom, 'sizes': self.splitter.sizes()}

    @staticmethod
    def close():
        if hasattr(ScriptEditor, 'gui'):
            QtWidgets.QMainWindow.close(ScriptEditor.gui)

    
if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    ScriptEditor.show()
    app.exec_()


logger.debug("Completed 'reading app/script_editor.py'")


# -*- coding: utf-8 -*-
from ..logger import logger
logger.debug("Started 'reading process/math_.py'")
import numpy as np
from qtpy import QtGui, QtWidgets, QtCore
from .. import global_vars as g
from ..utils.BaseProcess import BaseProcess, CheckBox, ComboBox
from ..window import Window

__all__ = ['subtract','multiply','divide','power','ratio','absolute_value','subtract_trace','divide_trace', 'sqrt']


def upgrade_dtype(dtype):
    if dtype==np.uint8:
        return np.uint16
    elif dtype==np.uint16:
        return np.uint32
    elif dtype==np.uint32:
        return np.uint64
    elif dtype==np.uint64:
        return np.int8
    elif dtype==np.int8:
        return np.int16
    elif dtype==np.int16:
        return np.int32
    elif dtype==np.int32:
        return np.int64
    elif dtype==np.float16:
        return np.float32
    elif dtype==np.float32:
        return np.float64


class Subtract(BaseProcess):
    """ subtract(value, keepSourceWindow=False)

    This takes a value and subtracts it from the current window's image.
    
    Parameters:
        value (int): The number you are subtracting.
    Returns:
        newWindow
    """
    __url__ = ""
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        value=QtWidgets.QDoubleSpinBox()
        if g.win is not None:
            maxx = np.max(g.win.image) * 100
            minn = -1 * maxx # -np.max sometimes returns abnormal large value
            value.setRange(minn, maxx)
            value.setValue(np.min(g.win.image))
        self.items.append({'name':'value','string':'Value','object':value})
        self.items.append({'name':'preview','string':'Preview','object':CheckBox()})
        super().gui()
    def __call__(self,value,keepSourceWindow=False):
        self.start(keepSourceWindow)
        if hasattr(value,'is_integer') and value.is_integer():
            value = int(value)
        if np.issubdtype(self.tif.dtype,np.integer):
            ddtype=np.iinfo(self.tif.dtype)
            while np.min(self.tif)-value<ddtype.min or np.max(self.tif)-value>ddtype.max: # if we exceed the bounds of the datatype
                ddtype=np.iinfo(upgrade_dtype(ddtype.dtype))
            self.newtif=self.tif.astype(ddtype.dtype)-value
        else:
            self.newtif=self.tif-value
        self.newname=self.oldname+' - Subtracted '+str(value)
        return self.end()
    def preview(self):
        value=self.getValue('value')
        preview=self.getValue('preview')
        if preview:
            testimage=np.copy(g.win.image[g.win.currentIndex])
            testimage=testimage-value
            g.win.imageview.setImage(testimage,autoLevels=False)
        else:
            g.win.reset()
subtract=Subtract()

class Subtract_trace(BaseProcess):
    """ subtract_trace(keepSourceWindow=False)

    This takes the most recently plotted trace and subtracts it from each corresponding frame.

    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def __call__(self,keepSourceWindow=False):
        self.start(keepSourceWindow)
        trace=g.currentTrace.rois[-1]['roi'].getTrace()
        nDims=len(self.tif.shape)
        if nDims !=3:
            g.alert('Wrong number of dimensions')
            return self.end()
        if self.tif.shape[0]!=len(trace):
            g.alert('Wrong trace length')
            return self.end()
        self.newtif=np.transpose(np.transpose(self.tif)-trace)
        self.newname=self.oldname+' - subtracted trace'
        return self.end()
subtract_trace=Subtract_trace()

class Divide_trace(BaseProcess):
    """ divide_trace(keepSourceWindow=False)

    This takes the most recently plotted trace and divides each pixel in the current Window by its value.

    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def __call__(self,keepSourceWindow=False):
        self.start(keepSourceWindow)
        trace=g.currentTrace.rois[-1]['roi'].getTrace()
        nDims=len(self.tif.shape)
        if nDims !=3:
            g.alert('Wrong number of dimensions')
            return self.end()
        if self.tif.shape[0]!=len(trace):
            g.alert('Wrong trace length')
            return self.end()
        self.newtif=np.transpose(np.transpose(self.tif)/trace)
        self.newname=self.oldname+' - divided trace'
        return self.end()
divide_trace=Divide_trace()
    
class Multiply(BaseProcess):
    """ multiply(value, keepSourceWindow=False)

    This takes a value and multiplies it to the current window's image.
    
    Parameters:
        value (float): The number you are multiplying by.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        value=QtWidgets.QDoubleSpinBox()
        value.setRange(-2**64,2**64)
        self.items.append({'name':'value','string':'Value','object':value})
        self.items.append({'name':'preview','string':'Preview','object':CheckBox()})
        super().gui()
    def __call__(self,value,keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif = self.tif*value
        self.newname = self.oldname+' - Multiplied by '+str(value)
        return self.end()
    def preview(self):
        value=self.getValue('value')
        preview=self.getValue('preview')
        if preview:
            testimage=np.copy(g.win.image[g.win.currentIndex])
            testimage=testimage*value
            g.win.imageview.setImage(testimage,autoLevels=False)
        else:
            g.win.reset()
multiply=Multiply()


class Divide(BaseProcess):
    """ divide(value, keepSourceWindow=False)

    This takes a value and divides it to the current window's image.

    Parameters:
        value (float): The number you are dividing by.
    Returns:
        newWindow
    """

    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        value = QtWidgets.QDoubleSpinBox()
        value.setRange(-2 ** 64, 2 ** 64)
        self.items.append({'name': 'value', 'string': 'Value', 'object': value})
        self.items.append({'name': 'preview', 'string': 'Preview', 'object': CheckBox()})
        super().gui()

    def __call__(self, value, keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif = self.tif / value
        self.newname = self.oldname + ' - Divided by ' + str(value)
        return self.end()

    def preview(self):
        value = self.getValue('value')
        preview = self.getValue('preview')
        if preview:
            testimage = np.copy(g.win.image[g.win.currentIndex])
            testimage = testimage / value
            g.win.imageview.setImage(testimage, autoLevels=False)
        else:
            g.win.reset()


divide = Divide()

class Power(BaseProcess):
    """ power(value, keepSourceWindow=False)

    This raises the current window's image to the power of 'value'.
    
    Parameters:
        value (int): The exponent.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        value=QtWidgets.QDoubleSpinBox()
        if g.win is not None:
            value.setRange(-100,100)
            value.setValue(1)
        self.items.append({'name':'value','string':'Value','object':value})
        self.items.append({'name':'preview','string':'Preview','object':CheckBox()})
        super().gui()
    def __call__(self,value,keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif=self.tif**value
        self.newname=self.oldname+' - Power of '+str(value)
        return self.end()
    def preview(self):
        value=self.getValue('value')
        preview=self.getValue('preview')
        if preview:
            testimage=np.copy(g.win.image[g.win.currentIndex])
            testimage=testimage**value
            g.win.imageview.setImage(testimage,autoLevels=False)
        else:
            g.win.reset()
power=Power()

class Sqrt(BaseProcess):
    """ sqrt(value, keepSourceWindow=False)

    This takes the square root of the current window's image.
    In this function, the square root of a negative number is set to 0.
    
    Parameters:
        value (int): The exponent.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        self.items.append({'name':'preview', 'string':'Preview', 'object':CheckBox()})
        super().gui()
    def __call__(self, keepSourceWindow=False):
        self.start(keepSourceWindow)
        A = np.copy(self.tif)
        A[A<0] = 0
        self.newtif = np.sqrt(A)
        self.newname = self.oldname+' - Sqrt '
        return self.end()
    def preview(self):
        preview = self.getValue('preview')
        if preview:
            A = np.copy(g.win.image[g.win.currentIndex])
            A[A<0] = 0
            A = np.sqrt(A)
            g.win.imageview.setImage(A, autoLevels=False)
        else:
            g.win.reset()
sqrt = Sqrt()
    
class Ratio(BaseProcess):
    """ ratio(first_frame, nFrames, ratio_type, black_level, keepSourceWindow=False)

    Takes a set of frames, combines them into a 2D array according to ratio_type, and divides the entire 3D array frame-by-frame by this array.
    
    Parameters:
        first_frame (int): The first frame in the set of frames to be combined
        nFrames (int): The number of frames to be combined.
        ratio_type (str): The method used to combine the frames.  Either 'standard deviation' or 'average'.
        black_level (float): The value to subtract from the entire movie prior to the ratio operation.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        nFrames = 1
        if g.win is not None:
            nFrames = g.win.image.shape[0]
        first_frame = QtWidgets.QSpinBox()
        first_frame.setMaximum(nFrames)
        self.items.append({'name': 'first_frame', 'string': 'First Frame', 'object': first_frame})
        nFrames_spinbox = QtWidgets.QSpinBox()
        nFrames_spinbox.setMaximum(nFrames)
        nFrames_spinbox.setMinimum(1)
        self.items.append({'name': 'nFrames', 'string': 'Number of Frames', 'object': nFrames_spinbox})
        ratio_type = ComboBox()
        ratio_type.addItem('average')
        ratio_type.addItem('standard deviation')
        self.items.append({'name': 'ratio_type', 'string': 'Ratio Type', 'object': ratio_type})
        black_level = QtWidgets.QDoubleSpinBox()
        black_level.setMinimum(0)
        black_level.setMaximum(1000)
        self.items.append({'name': 'black_level', 'string': 'Black Level', 'object': black_level})
        super().gui()

    def __call__(self, first_frame, nFrames, ratio_type, black_level=0, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.oldwindow.volume is None:
            A = self.tif - black_level
        else:
            A = self.oldwindow.volume - black_level
        if ratio_type == 'average':
            baseline = A[first_frame:first_frame+nFrames].mean(0)
            baseline[baseline == 0] = np.min(np.abs(baseline[baseline != 0])) # This isn't mathematically correct.  I do this to avoid dividing by zero
        elif ratio_type == 'standard deviation':
            baseline = A[first_frame:first_frame+nFrames].std(0)
            baseline[baseline == 0] = np.min(np.abs(baseline[baseline != 0]))
        else:
            g.alert("'{}' is an unknown ratio_type.  Try 'average' or 'standard deviation'".format(ratio_type))
            return None

        newA = (A/baseline).astype(g.settings['internal_data_type'])
        if self.oldwindow.volume is None:
            self.newtif=newA
            self.newname=self.oldname+' - Ratioed by '+str(ratio_type)
            return self.end()
        else:
            from plugins.light_sheet_analyzer.light_sheet_analyzer import Volume_Viewer
            self.newtif = np.squeeze(newA[:, 0, :, :])
            self.newname = self.oldname + ' - Ratioed by ' + str(ratio_type)
            w = self.end()
            w.volume = newA
            Volume_Viewer(w)
            return w
ratio=Ratio()


class Absolute_value(BaseProcess):
    """ absolute_value(keepSourceWindow=False)

    Takes the absolute value of a set of frames

    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        if super().gui()==False:
            return False
    def __call__(self,keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif=np.abs(self.tif)
        self.newname=self.oldname+' - Absolute Value'
        return self.end()
absolute_value=Absolute_value()


logger.debug("Completed 'reading process/math_.py'")


from qtpy import QtCore, QtGui, QtWidgets, PYQT5

__all__ = ['get_qapp']

qapp = None


def get_qapp(icon_path=None):
    """Get the QApplication instance currently in use. If no QApplication exists,
    one is created and the standard windoe icon is set to icon_path

    Args:
        icon_path (str): location of icon to use as default window icon

    Returns:
        QtGui.QApplication: the current application process 
    """
    global qapp
    qapp = QtWidgets.QApplication.instance()
    if qapp is None:
        qapp = QtWidgets.QApplication([''])
        qapp.setQuitOnLastWindowClosed(True)
        if icon_path is not None:
            qapp.setWindowIcon(QtGui.QIcon(icon_path))

    # Make sure we use high resolution icons with PyQt5 for HDPI
    # displays. TODO: check impact on non-HDPI displays.
    if PYQT5:
        qapp.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

    return qapp